--[[
**********************************************************************
MagicDKP Bidder - Addon to handle the DKP Bidding process
**********************************************************************
This file is part of MagicDKP, a World of Warcraft Addon

MagicDKP is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MagicDKP is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MagicDKP.  If not, see <http://www.gnu.org/licenses/>.
**********************************************************************
]]

local debug = false
local FAKE_LOOT_DEBUG = false
--[===[@debug@
debug = true
--FAKE_LOOT_DEBUG = true  -- this stops looting, and general fakes stuff for debugging of MagicDKP
--@end-debug@]===]

local MODULE_NAME = "Bidder"
local MODULE_VERSION = tonumber(("$Revision: 205 $"):match("%d+"))

local mod = MagicDKP

local module = mod.modules[MODULE_NAME] or mod:NewModule(MODULE_NAME, "AceEvent-3.0", "AceTimer-3.0")

if module.VERSION and module.VERSION < MODULE_VERSION then
   -- This will avoid overwriting an existing instance if it's newer. 
   -- Typical scenario is that someone has both MagicDKP_Bidder and MagicDKP/Bidder/
   -- at the same time.
   return
end

module.VERSION = MODULE_VERSION

if debug then
   BD = module
end
local MagicComm = LibStub("MagicComm-1.0")
local GUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("MagicDKP", false)
local tsort = table.sort
local fmt = string.format
local match = string.match
local GetNumRaidMembers = GetNumRaidMembers
local playerName
local prioSortVal = {}

module.defaults = {
   bidTimeout = 60,
   bidMessage = L["Attention: Bidding for [item] now open for [timeout] seconds."],
   bidMessageNoTime = L["Attention: Bidding for [item] now open."],
   bidClosedMessage = L["Bidding for [item] is now closed."],
   allowOverwrite = true,
   allowLowerBid = true,
   autoClose = false,
   hiddenBids = true,
   lootConfirmThreshold  = 3,
   autoBidThreshold  = 4,
   confirmAutoBid = true,
   confirmBankLoot = true,
   confirmDisenchantLoot = true,
   bidAmountType = 1,
   enableWhisperBids = true,
   enableMagicDKPClient = true,
   priorities = { "Primary Role", "Secondary Role", "Off Spec" },
   bidChannels = { CHAT_MSG_WHISPER = true  },
}

local lootThresholds = {
   [2] = ITEM_QUALITY2_DESC, 
   [3] = ITEM_QUALITY3_DESC, 
   [4] = ITEM_QUALITY4_DESC, 
   [5] = ITEM_QUALITY5_DESC
}

local db = {}
local info = {}

local function clear(mm) for id in pairs(mm) do mm[id] = nil end return mm end
local function count(hash) local count = 0 for i in pairs(hash) do count = count + 1 end return count end

function module:OnInitialize()
   -- register defaults and get db storage from the mothership
   db = mod:GetModuleDatabase(MODULE_NAME, module)
   module.bids = {}
   module.players = {}
   module.pendingLootEvents = {}
   module.pendingBidItems = {}
   playerName = UnitName("player")
   module.masterLootCandidates = {}

   StaticPopupDialogs["MDKPBidRemove"] = {
      text = L["Are you sure you want remove %s from the list?"],
      button1 = YES,
      button2 = NO,
      whileDead = 1,
      hideOnEscape = 1,
      startDelay = 0,
      timeout = 0,
      OnAccept = function(self) module:DeleteBidRow("CONFIRM") end
   }
   StaticPopupDialogs["MDKPStartNewBid"] = {
      text = L["Do you want to start a new bid for %s?"],
      button1 = YES,
      button2 = NO,
      whileDead = 1,
      hideOnEscape = 1,
      startDelay = 0,
      timeout = 0,
      OnAccept = function(self) module:StartNewBid(module.newPendingBidItem) end
   }
   StaticPopupDialogs["MDKPMagicLooter"] = {
      text = L["You need MagicLooter installed to use the disenchant and bank buttons. You can find it on Curse.com."],
      button1 = OKAY,
      whileDead = 1,
      hideOnEscape = 1,
      startDelay = 0,
      timeout = 0,
   }
   StaticPopupDialogs["MDKPNoCandidate"] = {
      text = L["Didn't find %s on the list of valid master loot candidates. Can't assign loot."],
      button1 = OKAY,
      whileDead = 1,
      hideOnEscape = 1,
      startDelay = 0,
      timeout = 0,
   }

   StaticPopupDialogs["MDKPBidInProgress"] = {
      text = L["A bid is currently in progress. Please wait for the bid to finish to assign the item."],
      button1 = OKAY,
      whileDead = 1,
      hideOnEscape = 1,
      startDelay = 0,
      timeout = 0,
   }
   StaticPopupDialogs["MDKPCouldntFindItem"] = {
      text = L["%s is not available to be looted. Is master looting enabled and is the loot window open?"],
      button1 = OKAY,
      whileDead = 1,
      hideOnEscape = 1,
      startDelay = 0,
      timeout = 0,
   }
   StaticPopupDialogs["MDKPNoRecipient"] = {
      text = L["Couldn't give the item since no recipient was found."],
      button1 = OKAY,
      whileDead = 1,
      hideOnEscape = 1,
      startDelay = 0,
      timeout = 0,
   }

   StaticPopupDialogs["MDKPConfirmLoot"] = {
      text = CONFIRM_LOOT_DISTRIBUTION,
      button1 = YES,
      button2 = NO,
      OnAccept = function(self,data) module:ReallyGiveLoot(data) end,
      timeout = 0,
      hideOnEscape = 1,
   }

   module.bidFrame = module:CreateBidFrame()
   module.CreateBidFrame = nil -- of to the garbage collector you go
end

function module:OnEnable()
   module:RegisterEvent("RAID_ROSTER_UPDATE", "UpdatePlayers")
   module:ReallyUpdatePlayers()
end

function module:OnDisable()
   module:UnregisterEvent("LOOT_SLOT_CLEARED")
   module:UnregisterEvent("LOOT_OPENED")
   module:UnregisterEvent("LOOT_CLOSED")
   module:UnregisterEvent("RAID_ROSTER_UPDATE")
end

function module:CheckLoot(...)
   -- Build a list of all master loot candidates
   for ci = 1, 40 do
      local candidate = GetMasterLootCandidate(ci)
      if candidate then
	 module.masterLootCandidates[candidate] = ci
      end
   end
   wipe(module.pendingBidItems)
   for slot = 1, GetNumLootItems() do
      local link = GetLootSlotLink(slot)
      if link and select(3, GetItemInfo(link)) >= db.autoBidThreshold then
	 module.pendingBidItems[slot] = link
      end
   end
   module:CheckNextBidItem()
   module:CheckLootSlots()
end

function module:LootSlotCleared(event, slotId)
   if slotId then
      local data = module.pendingLootEvents[slotId]
      module.pendingLootEvents[slotId] = nil
      if data then    
	 mod:HandleLoot(data.recipient, data.itemid, 1, FAKE_LOOT_DEBUG, true, data.isBank, data.isDE, data.dkp)
      end
      module.pendingBidItems[slotId] = nil
      module:CheckNextBidItem()
   end
   module:CheckLootSlots()
   if not module.masterLootCurrentItemSlot then
      module.lootDone = true
      module.bidFrame.bidButton:SetDisabled(true)
   end
end

function module:CheckNextBidItem()
   local first, link = next(module.pendingBidItems)
   if not first then return end
   if module.currentBidItem then
      -- to avoid starting a new bid for an item currently being bid for. Should also
      -- prevent new bids for the second of a double item drop, i.e armor tokens.
      for slot, item in pairs(module.pendingBidItems) do
	 if mod:LinkToItemID(item) == mod:LinkToItemID(module.currentBidItem) then
	    -- currently have a bid for an item on the body
	    return 
	 end
      end
   end
   
   if db.confirmAutoBid or module.bidTimeout then
      module.newPendingBidItem = link
      StaticPopup_Show("MDKPStartNewBid", link)
   else
      module:StartNewBid(link)
   end
end

function module:CheckLootSlots()
   local hasMasterLoot =  next(module.masterLootCandidates) 
   module.masterLootCurrentItemSlot = nil
   for slot = 1, GetNumLootItems() do
      local link =  GetLootSlotLink(slot)
      -- Comparing Item ID's since... the links possibly might be different
      -- This would explain why looting multiple items didn't work
      if hasMasterLoot and module.currentBidItem and not module.masterLootCurrentItemSlot then
	 mod:debug("Master loot enabled, and has a current bid item. Checking for match for %s.", module.currentBidItem)
	 if link and mod:LinkToItemID(link) == mod:LinkToItemID(module.currentBidItem) then
	    mod:debug("  Found item at slot %d.", slot)
	    module.currentBidItem = link 
	    module.masterLootCurrentItemSlot = slot
	 end
      end
   end
end

function module:VerifyButtonClick(recipient)
   if module.bidTimeout then 
      StaticPopup_Show("MDKPBidInProgress")
      return false
   end
   if not module.masterLootCurrentItemSlot and not FAKE_LOOT_DEBUG then
      -- Item not found on list, scan loot slots again to see if we can find it
      module:CheckLootSlots()
      if not module.masterLootCurrentItemSlot then
	 StaticPopup_Show("MDKPCouldntFindItem", module.currentBidItem)
	 return false
      end
   end
   if not recipient then
      StaticPopup_Show("MDKPNoRecipient")
      return false
   end
   return true
end

function module:GetBidWinner()
   local c = module.scroll.children
   if #c == 1 then return end
   local children = {}
   
   for i = 2, #c do
      children[#children + 1] = c[i]
   end
   tsort(children, module.sortMethods.priority)
   local bid = module.bids[children[1].player:GetText()]
   if bid and bid.bid then 
      return bid 
   end
end

-- This returns the bid amount for the person below the given recipient
-- but only if the priority is the same
function module:GetNextBidder(recipient, priority)
   if not recipient or not priority then return end -- safeguard
   local c = module.scroll.children
   if #c == 1 then return end
   local children = {}
   
   for i = 2, #c do children[#children + 1] = c[i] end

   tsort(children, module.sortMethods.priority)

   for id,child in ipairs(children) do
      if child.player:GetText() == recipient then
	 local nextId = id + 1
	 if children[nextId] then
	    local bid = module.bids[children[nextId].player:GetText()]
	    if bid and bid.bid and bid.priority == priority then
	       return bid.bid
	    end
	 end
	 return
      end
   end
end

function module:ClearLoot(event)
   clear(module.masterLootCandidates)
   module.masterLootCurrentItemSlot = false
   for slotId, data in pairs(module.pendingLootEvents) do
      module.pendingLootEvents[slotId] = nil
      mod:HandleLoot(data.recipient, data.itemid, 1, FAKE_LOOT_DEBUG, true, data.isBank, data.isDE, data.dkp)
   end
end

function module:StartNewBid(item)
   local itemid = mod:LinkToItemID(item)
   local bidFrame = module.bidFrame   
   bidFrame.item:SetText(item, itemid)
   module.lootDone = nil
   module.numBids = 0
   module:BidCompleted(true)   
   
   module.currentBidItem = item
   module.widgets = module.widgets or {}
   clear(module.widgets)
   clear(module.bids)
   clear(prioSortVal)
   module:SendBidRequest(item)
   if db.enableWhisperBids then
      for channel, enabled in pairs(db.bidChannels) do
	 if enabled then
	    if channel == "CHAT_MSG_RAID" then
	       module:RegisterEvent("CHAT_MSG_RAID_LEADER", "HandleBidMessage")
	    end
	    module:RegisterEvent(channel, "HandleBidMessage")
	 end
      end
   end
   -- Clear old list entries
   for i = 2, #module.scroll.children do
      GUI:Release(module.scroll.children[i])
      module.scroll.children[i] = nil
   end
   for _,player in ipairs(module.players) do
      local widget = module:GetBidRowWidget(player,"---", L["Bid Pending"])
      module.scroll:AddChild(widget)
   end
   module.scroll:DoLayout()
   
   bidFrame.bidButton:SetText(L["End Bidding"])
   bidFrame.bidButton:SetDisabled(false)
   
   module.lastSorted = "priority"
   module.reverse = false
   module.selectedPlayer = nil
   module.selectedButton:SetDisabled(true)
   for id,prio in pairs(db.priorities) do prioSortVal[prio] = id end
   prioSortVal[L["Pass"]] = #db.priorities+10000
   prioSortVal[L["|cfffe1f1fCan't use the item|r"]] = #db.priorities+10001
   MagicComm:RegisterListener(module, "MD")
   module.bidFrame:SetStatusText(L["Bidding pending..."])
end

local bidRunningRemainingMessage = L["Bid in progress: %d seconds remaining. %d/%d bids."]
local bidRunningMessage = L["Bid in progress. %d/%d bids."]
local bidCompletedMessage = L["Bid completed. %d/%d bids."]

function module:BidTimerUpdate()
   if db.bidTimeout > 0 then
      local remaining = (module.bidTimeout or 0) - time()
      if remaining <= 0 then
	 module:BidCompleted()
      else
	 module.bidFrame:SetStatusText(bidRunningRemainingMessage:format(remaining, module.numBids, #module.players))
      end
   else
      module.bidFrame:SetStatusText(bidRunningMessage:format(module.numBids, #module.players))
   end
end

function module:BidCompleted(quiet)
   if not module.bidTimeout then return end
   for channel in pairs(db.bidChannels) do
      module:UnregisterEvent(channel)
   end
   module:UnregisterEvent("CHAT_MSG_RAID_LEADER")

   module.bidFrame:SetStatusText(bidCompletedMessage:format(module.numBids, #module.players))

   if not quiet then 
      info.item = module.currentBidItem
      mod:SendChatMessage(mod:tokenize(db.bidClosedMessage, info), "GROUP")
   end

   mod:SetNetworkData("DKPCLOSE", module.currentBidItem)
   mod:SendUrgentMessage("RAID")

   module:CancelTimer(module.bidTimer, true)
   module.bidTimer = nil
   module.bidTimeout = nil
   module.bidFrame.bidButton:SetText(L["Redo Bid"])
   module:ShowHiddenBids()
   module:CheckLootSlots()
   module:ScheduleTimer(MagicComm.UnregisterListener, 5, MagicComm, module, "MD")
end

function module:OnDKPBid(item, priorities, bidmaster, expirationTime)
   -- This is here so the timeout is properly set if there's some initial
   -- "send lag" due to other addon communcation. Due to lack of a "unified"
   -- time stamp any duration based bidding is going to be somewhat off-sync
   -- with the server. Waiting until the bidmaster sees the bid makes it
   -- somewhat more accurate
   if bidmaster ~= playerName then
      -- we didn't start the bid so do nothing. 
      return
   end
   
   -- Send bid start raid chat message
   info.item = item
   info.timeout = db.bidTimeout
   mod:SendChatMessage(mod:tokenize(db.bidTimeout > 0 and db.bidMessage or db.bidMessageNoTime, info), "RW")

   -- Set the timeout and start the timer if needed
   if db.bidTimeout > 0 then module.bidTimeout = time() + db.bidTimeout else module.bidTimeout = time() + 10000000 end
   if module.bidTimer then module:CancelTimer(module.bidTimer, true) end

   if db.bidTimeout > 0 then
      module.bidTimer = module:ScheduleRepeatingTimer("BidTimerUpdate", 1)
   else
      module:BidTimerUpdate()
   end
   module.bidFrame:Show()
   --   module.bg.frame:Show() -- hmm...
   module:CheckLootSlots()
end
   
function module:OnDKPResponse(bidder, bid, priority, item, whisper)
   local response = module.bids[bidder]
   if not bid then
      priority = #db.priorities+1
   elseif bid == -1 then
      priority = #db.priorities + 2
      whisper = L["|cfffe1f1fCan't use the item|r"]
      bid = nil
   end
   
   if response then
      if not db.allowOverwrite then return end -- already bid, don't allow it again
      if not db.allowLowerBid and bid and response.bid and bid < response.bid then return end -- we can't lower bids
      response.widget:SetBid(bid or "---")
   else
      local widget = module.widgets[bidder]
      if widget then
	 widget:SetBid(bid or "---")
      else
	 widget = module:GetBidRowWidget(bidder, bid or "---", "")
	 module.scroll:AddChild(widget)
      end
      response = { 
	 bidder = bidder,
	 widget = widget
      }
      module.bids[bidder] = response
      module.numBids = module.numBids + 1
   end
   response.message = db.priorities[priority] or whisper or L["Pass"]
   response.whisper = whisper
   response.bid = bid
   response.priority = priority
   response.widget:SetPriority(response.message)
   if db.hiddenBids then
      if not module.bids[playerName] then
	 response.widget:SetBid("(hidden)")
	 response.widget:SetPriority(L["Bid hidden pending own bid"])
      elseif bidder == playerName then
	 module:ShowHiddenBids()
      end
   end
   module:BidDialog_Sort("RESORT");
   if module.bidTimeout then
      local numPlayers = #module.players
      if db.autoClose and numPlayers ~= 0 and module.numBids >= numPlayers then
	 module:BidCompleted()
      elseif db.bidTimeout == 0 then
	 module:BidTimerUpdate()
      end
   else
      module.bidFrame:SetStatusText(bidCompletedMessage:format(module.numBids, #module.players))
   end
end

function module:ShowHiddenBids()
   for bidder, response in pairs(module.bids) do
      response.widget:SetPriority(response.message)
      response.widget:SetBid(response.bid or "---")
   end
end

local dkpdata = {}
function module:SendBidRequest(item)
   if db.enableMagicDKPClient then
      local DB = mod:Database()
      if DB then
	 dkpdata.min, dkpdata.max, dkpdata.avg, dkpdata.num = DB:GetItemBidSummary(item)
	 dkpdata.multipool = mod:GetMultiPool(item)
	 module.dkpdata = dkpdata
      else
	 module.dkpdata = nil
      end
      mod:SetNetworkData("DKPBID", item, db.priorities, db.bidTimeout and db.bidTimeout > 0 and db.bidTimeout, DB and dkpdata)
      mod:SendUrgentMessage("RAID")
   else
      -- Short circuit to start the bidding
      module:OnDKPBid(item, db.priorities, playerName)
   end
end

function module:DeleteBidRow(event, player)
   if event == "CONFIRM" and module.deleteBidRow then
      local w = module.widgets[module.deleteBidRow]
      if w then
	 local c = module.scroll.children
	 local found 
	 for id = 2, #c do
	    if not found and c[id] == w then
	       GUI:Release(c[id])
	       found = true
	    end
	    if found then
	       c[id] = c[id+1]
	    end
	 end
	 module.scroll:DoLayout()
	 module.bids[module.deleteBidRow] = nil
	 module.widgets[module.deleteBidRow] = nil
	 module.deleteBidRow = nil
	 if player == module.selectedPlayer then
	    module.selectedPlayer = nil
	    module.selectedButton:SetDisabled(true)
	 end
      end
   else
      module.deleteBidRow = player
      StaticPopup_Show("MDKPBidRemove", player)
   end
end

function module:SelectBidRow(event, player)
   local c = module.scroll.children
   local w = module.widgets[player]
   if module.selectedPlayer == player then
      w:SetSelected(nil)
      module.selectedPlayer = nil
      module.selectedButton:SetDisabled(true)
   else
      module.selectedPlayer = player
      module:CheckLootSlots()
      for id = 2,#c do
	 c[id]:SetSelected(c[id] == w)
      end
      module.selectedButton:SetDisabled(false)
   end
end
   
function module:GetBidRowWidget(player, bid, prio)
   local row = GUI:Create("MDKPRow")
   local _,class = UnitClass(player)
   row:SetData(player, bid, prio, class and RAID_CLASS_COLORS[class])
   row:SetCurrent("---")
   if module.dkpdata then
      local dkp = mod:GetPlayerCurrentDKP(player, module.dkpdata.multipool)
      if dkp then
	 row:SetCurrent(tostring(dkp))
      end
   end
   row:SetCallback("OnRowDelete", module.DeleteBidRow)
   row:SetCallback("OnRowClick", module.SelectBidRow)
   module.widgets[player] = row
   return row
end


local function findpattern(text, pattern, start)
   local a, b = string.find(text, pattern, start)
   if a and b then 
      return string.sub(text, a, b)
   end
end

function module:HandleBidMessage(channel, msg, sender)
   msg = msg:trim()
   local bid = tonumber(msg) -- try numeric only bids first
   if not bid then bid = findpattern(msg, " %d+ ") end -- "bid 100 offspec"
   if not bid then bid = findpattern(msg, " %d+$") end -- "I bid 100"
   if not bid then bid = findpattern(msg, "^%d+ ") end -- "100 offspec"
   if not bid then return end -- no match, ignoring
   bid = tonumber(bid)
   msg = L["Whisper: "]..msg
   prioSortVal[msg] = #db.priorities+1
   module:OnDKPResponse(sender, bid, #db.priorities+1, module.currentBidItem, msg)
end

module.sortMethods = {
   bid = function(row1, row2) return (tonumber(row1.bid:GetText()) or -1) > (tonumber(row2.bid:GetText()) or -1) end,  
   bidReverse = function(row1, row2) return (tonumber(row1.bid:GetText()) or -1) < (tonumber(row2.bid:GetText()) or -1) end, 
   current = function(row1, row2) return (tonumber(row1.current:GetText()) or -1) > (tonumber(row2.current:GetText()) or -1) end,  
   currentReverse = function(row1, row2) return (tonumber(row1.current:GetText()) or -1) < (tonumber(row2.current:GetText()) or -1) end, 
   player = function(row1, row2) return row1.player:GetText() < row2.player:GetText() end, 
   playerReverse = function(row1, row2) return row1.player:GetText() > row2.player:GetText() end, 
   priority = function(row1, row2)
		 local p1, p2 = row1.priority:GetText(), row2.priority:GetText()
		 if p1 == p2 then
		    return module.sortMethods.bid(row1, row2)
		 end
		 return (prioSortVal[p1] or #db.priorities+1) < (prioSortVal[p2] or #db.priorities+1)
	      end, 
   priorityReverse = function(row1, row2)
		 local p1, p2 = row1.priority:GetText(), row2.priority:GetText()
		 if p1 == p2 then
		    return module.sortMethods.bid(row1, row2)
		 end
		 return (prioSortVal[p1] or #db.priorities+1) > (prioSortVal[p2] or #db.priorities+1)
	      end, 
}

function module:Debug() 
   module:OnDKPResponse("Amimage", math.random(1000), math.random(#db.priorities))
   module:OnDKPResponse("Herbalism", math.random(1000), math.random(#db.priorities))
   module:OnDKPResponse("Morasia", math.random(1000), math.random(#db.priorities))
   module:OnDKPResponse("Micke", math.random(1000), math.random(#db.priorities))
   module:OnDKPResponse("Jonas", math.random(1000), math.random(#db.priorities))
end

function module:BidDialog_Sort(event, row)
   local header
   local children = {}
   if event ~= "RESORT" then
      if module.lastSorted == row then
	 module.reverse = not module.reverse
      else
	 module.reverse = nil
      end
      module.lastSorted = row or module.lastSorted
   else
      row = module.lastSorted
   end
   
   if module.reverse then row = row.."Reverse" end
   header = module.scroll.children[1]
   for i = 2, #module.scroll.children do
      children[#children + 1] = module.scroll.children[i]
   end
   tsort(children, module.sortMethods[row])
   for i = 1, #children do
      module.scroll.children[i+1] = children[i]
   end
   module.scroll:DoLayout()
end

-- Creates the DKP Bid Master UI
function module:CreateBidFrame()
   local frame = GUI:Create("Frame")
   frame:SetTitle(L["DKP Bid Master"])
   frame:SetLayout("Flow")
   frame.width = "fill"
   frame.frame:SetFrameStrata("HIGH")
   frame:SetWidth(560)
   frame:SetHeight(560)
   frame.frame:SetMinResize(560, 545)
   frame.frame:SetMaxResize(600, 545)

   local item = GUI:Create("MDKPItemWidget")
   item:SetWidth(340)
   frame.item = item
   frame:AddChild(item)

   local button = GUI:Create("Button")
   button:SetWidth(120)
   button:SetText(L["End Bidding"])
   button:SetCallback("OnClick", function() module:HandleBidButton() end)
   frame.bidButton = button
   frame:AddChild(button)

   local scroll = GUI:Create("ScrollFrame")
   scroll:SetLayout("List")
   scroll:SetHeight(340)
   scroll.width = "fill"
   local header = GUI:Create("MDKPHeader")
   header:SetData(1, L["Player"])
   header:SetData(2, L["Bid"])
   header:SetData(3, L["Current"])
   header:SetData(4, L["Bid Priority"])
   header:SetCallback("OnRowClick", module.BidDialog_Sort)
   
   scroll:AddChild(header)
   module.scroll = scroll
   
   local ig = GUI:Create("InlineGroup")
   ig.width = "fill"
   ig.height = "fill"
   ig:SetLayout("Flow")
   ig:AddChild(scroll)
   frame:AddChild(ig)

   
   local bg = GUI:Create("InlineGroup")
   bg.width = "fill"
   bg:SetTitle("Loot item to...")   
   bg:SetLayout("Flow")
   bg.border:SetBackdrop(nil)
   ig:AddChild(bg)
   module.bg = bg

   button = GUI:Create("Button")
   button:SetText(L["Highest Bidder"])
   button:SetWidth(130)
   button:SetCallback("OnClick", function() module:GiveLoot(module:GetBidWinner()) end)
   bg:AddChild(button)
   module.winnerButton = button

   button = GUI:Create("Button")
   button:SetText(L["Selected Player"])
   button:SetCallback("OnClick", function() module:GiveLoot(module.selectedPlayer) end)
   button:SetWidth(130)
   bg:AddChild(button)
   module.selectedButton = button
   
   button = GUI:Create("Button")
   button:SetText(L["Disenchanter"])
   button:SetWidth(120)
   button:SetCallback("OnClick", function() if not MagicLooter then StaticPopup_Show("MDKPMagicLooter") else module:GiveLoot("_disenchant") end end)
   bg:AddChild(button)   
   module.disenchantButton = button

   
   button = GUI:Create("Button")
   button:SetText(L["Banker"])
   button:SetWidth(80)
   button:SetCallback("OnClick", function() if not MagicLooter then StaticPopup_Show("MDKPMagicLooter") else module:GiveLoot("_bank") end end)
   bg:AddChild(button)   
   module.bankButton = button
   frame:Hide()
   return frame
end

function module:HandleBidButton()
   if module.bidTimeout then
      -- end bid
      module:BidCompleted() 
   elseif not module.lootDone and module.currentBidItem then
      -- restart bidding
      module:StartNewBid(module.currentBidItem)
   end
end

function module:GiveLoot(recipient)
   if not module:VerifyButtonClick(recipient) then return end   

   local mlc, isBank, isDE, dkp
   if type(recipient) == "table" then
      dkp = recipient.bid
      recipient = recipient.bidder
      mlc = MagicLooter:GetLootCandidateID(recipient)
   elseif recipient == "_bank" then
      isBank = true
      mlc = MagicLooter:GetBankLootCandidateID()
      recipient = mlc and GetMasterLootCandidate(mlc) or "banker"
   elseif recipient == "_disenchant" then
      isDE = true
      mlc = MagicLooter:GetDisenchantLootCandidateID()
      recipient = mlc and GetMasterLootCandidate(mlc) or "disenchanter"
   else
      local row = module.bids[recipient]
      if row then dkp = row.dkp end
      mlc = MagicLooter:GetLootCandidateID(recipient)	 
   end

   if not mlc and not FAKE_LOOT_DEBUG then
      StaticPopup_Show("MDKPNoCandidate", recipient)       
      return
   end

   -- This will set the DKP amount to the 2nd highest bid (below the selected player) + 1
   if db.bidAmountType == 2 and not (isDE or isBank) then
      local nextBid = module:GetNextBidder(recipient, module.bids[recipient] and module.bids[recipient].priority)
      if nextBid then
	 dkp = nextBid + 1
      else
	 dkp = 1
      end
   end

   local data = {
      item = module.currentBidItem,
      itemid = mod:LinkToItemID(module.currentBidItem), 
      recipient = recipient,
      isDE = isDE,
      isBank = isBank,
      dkp = dkp,
      slot = module.masterLootCurrentItemSlot,
      mlc = mlc
   }

   if select(3, GetItemInfo(data.itemid)) >= db.lootConfirmThreshold 
      and (not (isDE or isBank) -- normal loot
	   or  (isDE and db.confirmDisenchantLoot) -- disenchant with disenchant confirmation on
	      or  (isBank and db.confirmBankLoot)  -- bank with bank confirmation on
	) then
      local dialog = StaticPopup_Show("MDKPConfirmLoot", module.currentBidItem, recipient)
      if dialog then dialog.data = data end      
      return
   end 
   module:ReallyGiveLoot(data)
end

function module:ReallyGiveLoot(data)
   if data.isDE then
      module.bidFrame:SetStatusText(L["Item given to %s for disenchanting."]:format(data.recipient))
   elseif data.isBank then
      module.bidFrame:SetStatusText(L["Item given to %s for banking."]:format(data.recipient))
   elseif data.dkp then
      module.bidFrame:SetStatusText(L["Item given to %s for a bid of %d dkp."]:format(data.recipient, data.dkp))
   else
      module.bidFrame:SetStatusText(L["Item given to %s."]:format(data.recipient))
   end

    -- Call MagicDKP to log the loot
   if not FAKE_LOOT_DEBUG then
      module.pendingLootEvents[data.slot] = data
      GiveMasterLoot(data.slot, data.mlc)
   else
      mod:HandleLoot(data.recipient, data.itemid, 1, FAKE_LOOT_DEBUG, true, data.isBank, data.isDE, data.dkp)
   end
end


do
   local _timer
   function module:UpdatePlayers()
      module:CancelTimer(_timer, true)
      _timer = module:ScheduleTimer("ReallyUpdatePlayers", 10)
   end
   
   function module:ReallyUpdatePlayers()
      local players, id, name, class, className, status = module.players
      clear(players)
      local isEnabled = mod:IsLootMasterAndEnabled()
      if isEnabled then 
	 for id = 1,GetNumRaidMembers() do
	    name, _, _, _, className, class, status = GetRaidRosterInfo(id)
	    if name and status ~= PLAYER_OFFLINE then 
	       players[#players+1]  = name
	    end
	 end
      end
      if #players > 0 and isEnabled then
	 if not module.hasEventsEnabled then
	    -- We're in a raid, and with events enabled in the main addon
	    if mod.hasDebug then
	       mod:debug("Bidder: Raid detected, enabling events.")
	    end
	    module:RegisterEvent("LOOT_OPENED", "CheckLoot")
	    module:RegisterEvent("LOOT_CLOSED", "ClearLoot")
	    module:RegisterEvent("LOOT_SLOT_CLEARED", "LootSlotCleared")
	 end
      elseif module.hasEventsEnabled then
	 if mod.hasDebug then
	    mod:debug("Bidder: No raid detected, disabling events.")
	 end
	 module:UnregisterEvent("LOOT_OPENED")
	 module:UnregisterEvent("LOOT_CLOSED")
	 module:UnregisterEvent("LOOT_SLOT_CLEARED")
      end
      module.hasEventsEnabled = #players > 0
   end
end


function module:SetProfileParam(var, value, sub)
   local varName = var[#var]
   if sub ~= nil then
      if type(db[varName]) ~= "table" then db[varName] = mod.get() end
      db[varName][value] = sub
   else
      if varName == "priorities" then
	 local newValues = db.varName or {}
	 clear(newValues)
	 for _,prio in ipairs(  { strsplit("\n", value) } ) do 
	    prio = prio:trim()
	    if strlen(prio) > 0 then
	       newValues[#newValues + 1] = prio
	    end
	 end
	 value = newValues
      end
      db[varName] = value
   end
end

function module:GetProfileParam(var, sub)
   local varName = var[#var]
   if sub ~= nil then
      return type(db[varName]) == "table" and db[varName][sub]
   elseif varName == "priorities" then
      if type(db[varName]) == "table" then 
	 return strjoin("\n", unpack(db[varName]))
      else
	 return ""
      end
   else
      return db[varName]
   end
end

function module:OnProfileChanged(newdb)
   db = newdb
end

-- config options
module.options = {
   type = "group",
   name = L["DKP Bidder"],
   handler = module,
   childGroups  = "tab",
   set = "SetProfileParam",
   get = "GetProfileParam", 
   args = {
      options = {
	 type = "group",
	 name = L["General"],
	 order = 1,
	 args = {
	    bidTimeout = {
	       type = "range",
	       name = L["Bid Timeout"],
	       desc = L["Time in seconds before the bid time expires. Set to zero to avoid bid expiration."],
	       min = 0, max = 120, step = 1,
	       order = 20,
	       width="full",
	    },
	    bidExtension = {
	       type = "range",
	       name = L["Bid Extension"],
	       desc = L["Minimum remaining duration of the bidding after receiving a new bid. This can be used to automatically extend the bid duration after a bid is received. A value of zero disables the feature."],
	       min = 0, max = 60, step = 1,
	       order = 21,
	       hidden = true,
	       width = "full", 
	    },
	    autoClose = {
	       type = "toggle",
	       name = L["Close Bid Automatically"],
	       desc = L["When enabled, bidding will be closed early if everyone on the bid list has entered a bid or passed. Passing currently depends on MagicDKP Client - there's no passing mechanism in whisper bidding."],
	       width = "full",
	       order = 70,
	    },
	    enableWhisperBids = {
	       type = "toggle",
	       name = L["Enable Whisper Bids"],
	       desc = L["When enabled MagicDKP Bidder will listen to bids sent via chat messages in one or more of raid chat, whispers or says."],
	       width = "full",
	       order = 70,
	    },
	    enableMagicDKPClient = {
	       type = "toggle",
	       name = L["Enable MagicDKP Client Bids"],
	       desc = L["When enabled MagicDKP Bidder will send bid events to MagicDKP Client."],
	       width = "full",
	       order = 70,
	    },
	    priorities = {
	       type = "input",
	       multiline = true,
	       name = L["Bid Priorities"],
	       desc = L["The list of bid priorities that are sent to the MagicDKP Client. Useful if you separate bids main spec and off spec bids for example. Has no effect on whisper bids."],
	       width = "full",
	       order = 1
	    },
	    bidAmountType = {
	       type = "select",
	       name = L["Bid Amount Method"],
	       desc = L["The method used when determining how much dkp a user should pay. The methods are: Bid Amount (the exact amount the user entered), Second Highest +1 (second highest bid amount plus 1 dkp)."],
	       values = {
		  L["Bid Amount"], L["Second Highest +1"]
	       },
	       order = 100, 
	    },
	    autoBidThreshold = {
	       type = "select",
	       name = L["Auto Bid Quality Threshold"],
	       desc = L["When opening a loot windows, query for bidding automatically for items at or above this item quality."],
	       values = lootThresholds,
	       order = 102,
	    },
	    
	    confirmAutoBid = {
	       type = "toggle",
	       name = L["Confirm New Auto Bid"],
	       desc = L["When enabled, you'll be asked to confirm each auto started bid. Otherwise they are started without player interaction."],
	       order = 103
	    }
	 }
      },
      confirmation = {
	 type = "group",
	 name = L["Loot Confirmation"],
	 args = {
	    lootConfirmThreshold = {
	       type = "select",
	       name = L["Loot Confirmation Threshold"],
	       desc = L["When giving loot of this threshold or higher, ask for confirmation."],
	       values = lootThresholds, 
	    }, 
	    confirmBankLoot = {
	       type = "toggle",
	       width = "full", 
	       name = L["Confirm Bank Loot"],
	       desc = L["Ask for confirmation when sending loot to the bank."],
	    }, 
	    confirmDisenchantLoot = {
	       type = "toggle",
	       width = "full", 
	       name = L["Confirm Disenchant Loot"],
	       desc = L["Ask for confirmation when sending loot to be disenchanted."],
	       order = 1
	    },
	    spacer = {
	       type = "description",
	       name = "",
	       order = 2
	    }
	 }
      },
      messaging = {
	 type = "group",
	 name = L["Messages"],
	 args = {
	    bidMessage = {
	       type = "input",
	       width="full",
	       name = L["Bid Announce Message With Duration"],
	       desc = L["The message sent to the raid or party when a new bid begins. Used for bids with a duration. The following token can be used: [item] (the item link) and [timeout] (the bid timeout)."],
	       order = 40,
	       disabled = function() return  not db.bidTimeout or db.bidTimeout == 0 end
	    },
	    bidMessageNoTime = {
	       type = "input",
	       width="full",
	       name = L["Bid Announce Message Without Duration"],
	       desc = L["The message sent to the raid or party when a new bid begins. Used when the bid has no duretion set. The following token can be used: [item] (the item link)"],
	       order = 40,
	       disabled = function() return db.bidTimeout and db.bidTimeout > 0 end
	    },
	    bidClosedMessage = {
	       type = "input",
	       width="full",
	       name = L["Bid Close Message"],
	       desc = L["The message sent to the raid or party when the bidding is closed. The following token can be used: [item]."],
	       order = 42,
	    },
	 },
      },
      whisper = {
	 type = "group",
	 name = L["Whisper Bidding"],
	 disabled = function() return not db.enableWhisperBids end,
	 args = {	 
	    bidChannels = {
	       type = "multiselect",
	       values =  {
		  CHAT_MSG_WHISPER = L["Whisper"],
		  CHAT_MSG_RAID = L["Raid"],
		  CHAT_MSG_SAY = L["Say"], 
	       },
	       name = L["Bid Channels"],
	       desc = L["The channels that DKPBidder will listen to bids on. If no channels are selected, only bids sent by MagicDKP_Client will be recorded."], 
	       order = 50, 
	    },
	    allowOverwrite = {
	       type = "toggle",
	       name = L["Allow Bid Changes"],
	       desc = L["If enabled whisper bids are allowed to overrwrite previous bids. "], 
	       width = "full",
	       order = 60, 
	    },
	    allowLowerBid= {
	       type = "toggle",
	       name = L["Allow Lowering Bids"],
	       desc = L["If enabled whisper bids are allowed to lower a previous bid. "],
	       hidden = function() return not db.allowOverwrite end, 
	       width = "full", 
	       order = 60, 
	    },
	 },
      }, 
   }
}

